from jolt import influence
from jolt import Task
from jolt import utils
from jolt.error import raise_task_error_if

import os


class CMake(Task):
    """ Builds and publishes a CMake project """

    abstract = True

    cmakelists = "CMakeLists.txt"
    """ Path to CMakeLists.txt or directory containing CMakelists.txt """

    generator = None

    incremental = True

    options = []
    """ List of options and their values (``option[:type]=value``) """

    def run(self, deps, tools):
        raise_task_error_if(not self.cmakelists, self, "cmakelists attribute has not been defined")

        cmake = tools.cmake(incremental=self.incremental)
        cmake.configure(tools.expand(self.cmakelists), *["-D" + tools.expand(option) for option in self.options], generator=self.generator)
        cmake.build()
        cmake.install()

    def publish(self, artifact, tools):
        cmake = tools.cmake()
        cmake.publish(artifact)


@influence.attribute("headers", type=influence.FileInfluence)
@influence.attribute("sources", type=influence.FileInfluence)
class _CMakeCXX(CMake):
    asflags = []
    binary = None
    cflags = []
    cxxflags = []
    headers = []
    incpaths = []
    ldflags = []
    macros = []
    shared = False
    sources = []

    generator = None
    incremental = True
    template = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.binary = self.binary or self.canonical_name

    def run(self, deps, tools):
        with tools.cwd(tools.builddir(incremental=self.incremental)):
            self.cmakelists = tools.expand_path(self.cmakelists)
            project = utils.render(self.template, deps=deps, task=self, tools=tools, os=os)
            tools.write_file(self.cmakelists, project, expand=False)
        super().run(deps, tools)

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.headers:
            for header in self.headers:
                artifact.collect(header)
            artifact.cxxinfo.incpaths.append(".")


class CXXLibrary(_CMakeCXX):
    abstract = True
    template = "cxxlibrary.cmake.template"

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        if self.headers:
            for header in self.headers:
                artifact.collect(header)
            artifact.cxxinfo.incpaths.append(".")
        artifact.cxxinfo.libraries.append("{binary}")
        artifact.cxxinfo.libpaths.append("lib")


class CXXExecutable(_CMakeCXX):
    abstract = True
    template = "cxxexecutable.cmake.template"

    def publish(self, artifact, tools):
        super().publish(artifact, tools)



from jolt import cache
from jolt import cli
from jolt import config
from jolt import graph
from jolt import log
from jolt import scheduler
from jolt.manifest import JoltManifest
from jolt.options import JoltOptions
from jolt.tasks import TaskRegistry
from jolt.tools import Tools

import click
import getpass
import keyring
import requests
import urllib3
import json

urllib3.disable_warnings()


NAME = "snap"


def _get_auth():
    service = config.get(NAME, "keyring.service", "jenkins")
    username = config.get(NAME, "keyring.username") or \
        input("Jenkins username: ")
    if username:
        config.set(NAME, "keyring.username", username)
        config.save()
    password = config.get(NAME, "keyring.password") or \
        keyring.get_password(service, username)
    if not password:
        password = getpass.getpass("Jenkins password: ")
        assert password, "no password in keyring for " + service
        keyring.set_password(service, username, password)
    return username, password


@cli.cli.command(name="snap")
@click.argument("task", type=str, required=True, shell_complete=cli._autocomplete_tasks)
@click.option("-d", "--default", type=str, multiple=True, help="Override default parameter values.")
@click.pass_context
def snap(ctx, task, default):
    """ Submit a task for background build in a Jenkins job """

    manifest = ctx.obj["manifest"]
    options = JoltOptions(default=default)
    acache = cache.ArtifactCache.get(options)
    registry = TaskRegistry.get()

    for params in default:
        registry.set_default_parameters(params)

    gb = graph.GraphBuilder(registry, acache, manifest, options, progress=True)
    dag = gb.build([task])

    manifest = JoltManifest.export(dag.requested_goals)
    manifest.config = ""
    build = manifest.create_build()

    for task in dag.goals:
        mt = build.create_task()
        mt.name = task.qualified_name

    for task in options.default:
        default = build.create_default()
        default.name = task

    registry = scheduler.ExecutorRegistry.get()
    for key, value in registry.get_network_parameters(dag.requested_goals[0]).items():
        param = manifest.create_parameter()
        param.key = key
        param.value = value

    url = config.get("snap", "url")

    jolt_id = manifest.get_parameter("jolt_identity")
    jolt_url = manifest.get_parameter("jolt_url")

    params = {
        "parameter": [
            {
                "name": "default.joltxmanifest",
                "file": "file0",
            },
            {
                "name": "jolt.config",
                "file": "file1",
            },
            {
                "name": "JOLT_TASK",
                "value": dag.requested_goals[0].qualified_name,
            },
            {
                "name": "JOLT",
                "value": jolt_id,
            },
            {
                "name": "JOLT_URL",
                "value": jolt_url,
            },
            {
                "name": "JOLT_NOTIFY",
                "value": config.get(NAME, "notify", ""),
            },
        ]
    }

    tools = Tools()
    data, content_type = urllib3.encode_multipart_formdata([
        ("file0", ("default.joltxmanifest", manifest.format())),
        ("file1", ("jolt.config", tools.read_file(config.location))),
        ("json", json.dumps(params)),
        ("Submit", "Build"),
    ])

    resp = requests.post(url + "/build", auth=_get_auth(), data=data,
                         headers={"content-type": content_type}, verify=False)
    resp.raise_for_status()
    log.info("Build successfully submitted")



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from jolt.plugins.remote_execution import administration_pb2 as jolt_dot_plugins_dot_remote__execution_dot_administration__pb2
from jolt.plugins.remote_execution import scheduler_pb2 as jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2


class AdministrationStub(object):
    """This service is not yet implemented.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CancelBuild = channel.unary_unary(
                '/Administration/CancelBuild',
                request_serializer=jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2.CancelBuildRequest.SerializeToString,
                response_deserializer=jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2.CancelBuildResponse.FromString,
                )
        self.ListBuilds = channel.unary_unary(
                '/Administration/ListBuilds',
                request_serializer=jolt_dot_plugins_dot_remote__execution_dot_administration__pb2.ListBuildsRequest.SerializeToString,
                response_deserializer=jolt_dot_plugins_dot_remote__execution_dot_administration__pb2.ListBuildsResponse.FromString,
                )
        self.ListWorkers = channel.unary_unary(
                '/Administration/ListWorkers',
                request_serializer=jolt_dot_plugins_dot_remote__execution_dot_administration__pb2.ListWorkersRequest.SerializeToString,
                response_deserializer=jolt_dot_plugins_dot_remote__execution_dot_administration__pb2.ListWorkersResponse.FromString,
                )
        self.Reschedule = channel.unary_unary(
                '/Administration/Reschedule',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )


class AdministrationServicer(object):
    """This service is not yet implemented.
    """

    def CancelBuild(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListBuilds(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListWorkers(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Reschedule(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AdministrationServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CancelBuild': grpc.unary_unary_rpc_method_handler(
                    servicer.CancelBuild,
                    request_deserializer=jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2.CancelBuildRequest.FromString,
                    response_serializer=jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2.CancelBuildResponse.SerializeToString,
            ),
            'ListBuilds': grpc.unary_unary_rpc_method_handler(
                    servicer.ListBuilds,
                    request_deserializer=jolt_dot_plugins_dot_remote__execution_dot_administration__pb2.ListBuildsRequest.FromString,
                    response_serializer=jolt_dot_plugins_dot_remote__execution_dot_administration__pb2.ListBuildsResponse.SerializeToString,
            ),
            'ListWorkers': grpc.unary_unary_rpc_method_handler(
                    servicer.ListWorkers,
                    request_deserializer=jolt_dot_plugins_dot_remote__execution_dot_administration__pb2.ListWorkersRequest.FromString,
                    response_serializer=jolt_dot_plugins_dot_remote__execution_dot_administration__pb2.ListWorkersResponse.SerializeToString,
            ),
            'Reschedule': grpc.unary_unary_rpc_method_handler(
                    servicer.Reschedule,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Administration', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Administration(object):
    """This service is not yet implemented.
    """

    @staticmethod
    def CancelBuild(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Administration/CancelBuild',
            jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2.CancelBuildRequest.SerializeToString,
            jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2.CancelBuildResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListBuilds(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Administration/ListBuilds',
            jolt_dot_plugins_dot_remote__execution_dot_administration__pb2.ListBuildsRequest.SerializeToString,
            jolt_dot_plugins_dot_remote__execution_dot_administration__pb2.ListBuildsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListWorkers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Administration/ListWorkers',
            jolt_dot_plugins_dot_remote__execution_dot_administration__pb2.ListWorkersRequest.SerializeToString,
            jolt_dot_plugins_dot_remote__execution_dot_administration__pb2.ListWorkersResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Reschedule(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Administration/Reschedule',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from jolt import BooleanParameter, Parameter, Task
from jolt import attributes, filesystem as fs, loader
from jolt.tasks import TaskRegistry

from jolt.cache import ArtifactListAttribute
from jolt.cache import ArtifactAttributeSet
from jolt.cache import ArtifactAttributeSetProvider

import contextlib
import os
import platform


class DebianListVariable(ArtifactListAttribute):
    pass


class DebianAttributeSet(ArtifactAttributeSet):
    def __init__(self, artifact):
        super(DebianAttributeSet, self).__init__()
        super(ArtifactAttributeSet, self).__setattr__("_artifact", artifact)

    def create(self, name):
        if name == "chroot":
            return DebianListVariable(self._artifact, "chroot")
        assert False, "No such debian attribute: {0}".format(name)


@ArtifactAttributeSetProvider.Register
class DebianAttributeProvider(ArtifactAttributeSetProvider):
    def create(self, artifact):
        setattr(artifact, "debian", DebianAttributeSet(artifact))

    def parse(self, artifact, content):
        if "debian" not in content:
            return
        for key, value in content["debian"].items():
            getattr(artifact.debian, key).set_value(value, expand=False)

    def format(self, artifact, content):
        if "debian" not in content:
            content["debian"] = {}
        for key, attrib in artifact.debian.items():
            content["debian"][key] = attrib.get_value()

    def apply_deps(self, task, deps):
        lowerdirs = []
        for _, artifact in deps.items():
            lowerdirs += [
                fs.path.join(artifact.path, item)
                for item in artifact.debian.chroot.items()
            ]
        if lowerdirs:
            task.__chroot = task.tools.builddir("chroot")
            with task.tools.cwd(task.__chroot):
                task.tools.mkdir("work")
                task.tools.mkdir("uppr")
                task.tools.mkdir("root")
                task.tools.run("fuse-overlayfs -o lowerdir={},upperdir=uppr,workdir=work root",
                               ":".join(lowerdirs))
            task.__unshare = contextlib.ExitStack()
            task.__unshare.enter_context(
                task.tools.chroot(chroot=fs.path.join(task.__chroot, "root")))

    def unapply_deps(self, task, deps):
        try:
            task.__unshare.close()
            with task.tools.cwd(task.__chroot):
                task.tools.run("fusermount -u root")
        except Exception:
            pass


class Debian(Task):
    abstract = False

    packages = []
    """ List of packages to install, in addition to those included by variant. """

    mode = "unshare"
    """
    Mode of operation when creating Debian chroot.

    Valid modes are:

      - fakeroot
      - proot
      - unshare

    The default mode is fakeroot. See mmdebstrap documentation
    for details. Note that Jolt does not permit any mode that requires
    root privileges.
    """

    mirrors = []
    """ List of mirrors """

    suite = "bullseye"
    """
    Debian release codename.

    The suite may be a valid release code name (eg, sid, stretch, jessie)
    or a symbolic name (eg, unstable, testing, stable, oldstable).
    Any suite name that works with apt on the given mirror will work.
    """

    target = "rootfs.tar"
    """ Target """

    variant = "extract"
    """
    Choose which package set to install.

    Valid variant names are:
      - extract
      - custom
      - essential
      - apt
      - required
      - minbase
      - buildd
      - important
      - debootstrap
      - standard

    The default variant is extract where only packages listed in ``packages``
    are downloaded and extracted, but not installed.

    See mmdebstrab documentation for details.
    """

    def run(self, deps, tools):
        with tools.cwd(tools.builddir()):
            packages = "--include={}".format(",".join(self.packages)) if self.packages else ""
            tools.run(
                "mmdebstrap {} --mode={mode} --variant={variant} {suite} {target}",
                packages,
            )

    def publish(self, artifact, tools):
        with tools.cwd(tools.builddir()):
            artifact.collect("*", symlinks=True)


class DebianPkgBase(Debian):
    abstract = True

    include = ["."]
    exclude = []

    def run(self, deps, tools):
        super().run(deps, tools)
        with tools.cwd(tools.builddir()):
            include = " ".join(self.include)
            exclude = " ".join(["--exclude=" + i for i in self.exclude])
            tools.run("tar --exclude=./dev {} -xf {target} {}", exclude, include)
            tools.unlink(self.target)

    def publish(self, artifact, tools):
        super().publish(artifact, tools)

        with tools.cwd(artifact.path):
            for bindir in ["bin", "sbin", "usr/bin", "usr/sbin"]:
                if os.path.exists(tools.expand_path(bindir)):
                    artifact.environ.PATH.append(bindir)

            for incdir in ["usr/include"]:
                if os.path.exists(tools.expand_path(incdir)):
                    pass  # artifact.cxxinfo.incpaths.append(incdir)

            for libdir in ["lib", "usr/lib"]:
                if os.path.exists(tools.expand_path(libdir)):
                    pass  # artifact.cxxinfo.libpaths.append(libdir)
                    artifact.environ.LD_LIBRARY_PATH.append(libdir)

            for pkgdir in ["usr/share/pkgconfig"]:
                if os.path.exists(tools.expand_path(pkgdir)):
                    artifact.environ.PKG_CONFIG_PATH.append(pkgdir)


@attributes.method("run", "run_{download[download,build]}")
class MMDebstrap(DebianPkgBase):
    """
    Provides the tools required to build Debian tools and filesystems.
    """

    name = "debian/mmdebstrap"
    packages = [
        "bash",
        "bubblewrap",
        "coreutils",
        "dash",
        "fakechroot",
        "fakeroot",
        "fuse-overlayfs",
        "mmdebstrap",
        "proot",
        "qemu-user",
        "util-linux",
    ]
    mode = "unshare"
    suite = "bullseye"
    variant = "essential"

    download = BooleanParameter(True, help="Download files from Github, or build from scratch.")
    url = "https://github.com/srand/jolt/releases/download/v0.9.35/mmdebstrap.tgz"

    @property
    def joltdir(self):
        return loader.JoltLoader.get().joltdir

    def run_build(self, deps, tools):
        super().run(deps, tools)

    def run_download(self, deps, tools):
        with tools.cwd(tools.builddir()):
            assert tools.download(self.url, "debstrap.tgz"), "Failed to download mmdebstrap"
            tools.extract("debstrap.tgz", ".")
            tools.unlink("debstrap.tgz")


TaskRegistry.get().add_task_class(MMDebstrap)


class DebianPkg(DebianPkgBase):
    abstract = True
    requires = ["debian/mmdebstrap"]

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.debian.chroot.append(".")


class DebianEssential(DebianPkg):
    name = "debian/essential"
    variant = "essential"

    @property
    def joltdir(self):
        return loader.JoltLoader.get().joltdir


TaskRegistry.get().add_task_class(DebianEssential)


def _host():
    m = platform.machine()
    if m == "x86_64":
        return "amd64"
    return m


@attributes.attribute("packages", "packages_{host}_{arch}")
@attributes.method("publish", "publish_{host}_{arch}")
class GCC(DebianPkg):
    name = "debian/gcc"

    host = Parameter(_host(), values=["amd64"], const=True)
    arch = Parameter(_host(), values=["amd64", "armel", "armhf", "arm64"], help="Target architecture.")

    packages_amd64_amd64 = ["gcc", "g++"]
    packages_amd64_armhf = ["crossbuild-essential-armhf"]
    packages_amd64_armel = ["crossbuild-essential-armel"]
    packages_amd64_arm64 = ["crossbuild-essential-arm64"]

    @property
    def joltdir(self):
        return loader.JoltLoader.get().joltdir

    def publish_amd64_amd64(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.AR = "ar"
        artifact.environ.CC = "gcc"
        artifact.environ.CONFIGURE_FLAGS = ""
        artifact.environ.CPP = "cpp"
        artifact.environ.CROSS_COMPILE = ""
        artifact.environ.CXX = "g++"
        artifact.environ.LD = "ld"
        artifact.environ.OBJCOPY = "objcopy"
        artifact.environ.OBJDUMP = "objdump"
        artifact.environ.PKG_CONFIG = "pkg-config"
        artifact.environ.RANLIB = "ranlib"
        artifact.environ.STRIP = "strip"
        artifact.environ.TARGET_PREFIX = ""

    def publish_amd64_armel(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.AR = "arm-linux-gnueabi-ar"
        artifact.environ.CC = "arm-linux-gnueabi-gcc"
        artifact.environ.CONFIGURE_FLAGS = "--target arm-linux-gnueabi --host arm-linux-gnueabi --build x86_64-linux-gnu"
        artifact.environ.CPP = "arm-linux-gnueabi-cpp"
        artifact.environ.CROSS_COMPILE = "arm-linux-gnueabi-"
        artifact.environ.CXX = "arm-linux-gnueabi-g++"
        artifact.environ.LD = "arm-linux-gnueabi-ld"
        artifact.environ.NM = "arm-linux-gnueabi-nm"
        artifact.environ.OBJCOPY = "arm-linux-gnueabi-objcopy"
        artifact.environ.OBJDUMP = "arm-linux-gnueabi-objdump"
        artifact.environ.PKG_CONFIG = "arm-linux-gnueabi-pkg-config"
        artifact.environ.RANLIB = "arm-linux-gnueabi-ranlib"
        artifact.environ.STRIP = "arm-linux-gnueabi-strip"
        artifact.environ.TARGET_PREFIX = "arm-linux-gnueabi-"

    def publish_amd64_armhf(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.AR = "arm-linux-gnueabihf-ar"
        artifact.environ.CC = "arm-linux-gnueabihf-gcc"
        artifact.environ.CONFIGURE_FLAGS = "--target arm-linux-gnueabihf --host arm-linux-gnueabihf --build x86_64-linux-gnu"
        artifact.environ.CPP = "arm-linux-gnueabihf-cpp"
        artifact.environ.CROSS_COMPILE = "arm-linux-gnueabihf-"
        artifact.environ.CXX = "arm-linux-gnueabihf-g++"
        artifact.environ.LD = "arm-linux-gnueabihf-ld"
        artifact.environ.NM = "arm-linux-gnueabihf-nm"
        artifact.environ.OBJCOPY = "arm-linux-gnueabihf-objcopy"
        artifact.environ.OBJDUMP = "arm-linux-gnueabihf-objdump"
        artifact.environ.PKG_CONFIG = "arm-linux-gnueabihf-pkg-config"
        artifact.environ.RANLIB = "arm-linux-gnueabihf-ranlib"
        artifact.environ.STRIP = "arm-linux-gnueabihf-strip"
        artifact.environ.TARGET_PREFIX = "arm-linux-gnueabihf-"

    def publish_amd64_arm64(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.AR = "aarch64-linux-gnu-ar"
        artifact.environ.CC = "aarch64-linux-gnu-gcc"
        artifact.environ.CONFIGURE_FLAGS = "--target aarch64-linux-gnu --host aarch64-linux-gnu --build x86_64-linux-gnu"
        artifact.environ.CPP = "aarch64-linux-gnu-cpp"
        artifact.environ.CROSS_COMPILE = "aarch64-linux-gnu-"
        artifact.environ.CXX = "aarch64-linux-gnu-g++"
        artifact.environ.LD = "aarch64-linux-gnu-ld"
        artifact.environ.NM = "aarch64-linux-gnu-nm"
        artifact.environ.OBJCOPY = "aarch64-linux-gnu-objcopy"
        artifact.environ.OBJDUMP = "aarch64-linux-gnu-objdump"
        artifact.environ.PKG_CONFIG = "aarch64-linux-gnu-pkg-config"
        artifact.environ.RANLIB = "aarch64-linux-gnu-ranlib"
        artifact.environ.STRIP = "aarch64-linux-gnu-strip"
        artifact.environ.TARGET_PREFIX = "aarch64-linux-gnu-"


TaskRegistry.get().add_task_class(GCC)



from jolt import MultiTask, attributes
from jolt import utils
from jolt.error import raise_task_error, raise_task_error_if
from jolt.influence import FileInfluence
from jolt.plugins import ninja

from copy import copy
import os


class GnuToolchain(object):
    gnu_ar = "ar"
    gnu_as = "as"
    gnu_cc = "gcc"
    gnu_cxx = "g++"
    gnu_ld = "g++"

    def _gnu_tool(self, env, default):
        cmd = self.tools.getenv(env, default).split()
        tool = self.tools.which(cmd[0])
        raise_task_error_if(not tool, self, f"Tool '{default}' not found in PATH")
        cmd[0] = tool
        return " ".join(cmd)

    gnu_archive_cmd = "{gnu_ar} -M <{inputs}"
    gnu_compile_asm_cmd = "{gnu_cc} -x assembler -MMD -MF {outputs}.d {asflags} -c {inputs} -o {outputs}"
    gnu_compile_asm_cpp_cmd = "{gnu_cc} -x assembler-with-cpp -MMD -MF {outputs}.d {% if shared %}-fPIC {% endif %}{asflags_imported} {asflags_environ} {asflags} {{(macros_imported+macros)|prefix('-D')|join(' ')}} {{(incpaths+incpaths_imported)|prefix('-I')|join(' ')}} -c {inputs} -o {outputs}"
    gnu_compile_c_cmd = "{gnu_cc} -x c -MMD -MF {outputs}.d {% if shared %}-fPIC {% endif %}{cflags_imported} {cflags_environ} {cflags} {{(macros_imported+macros)|prefix('-D')|join(' ')}} {{(incpaths+incpaths_imported)|prefix('-I')|join(' ')}} -c {inputs} -o {outputs}"
    gnu_compile_cxx_cmd = "{gnu_cxx} -x c++ -MMD -MF {outputs}.d {% if shared %}-fPIC {% endif %}{cxxflags_imported} {cxxflags_environ} {cxxflags} {{(macros_imported+macros)|prefix('-D')|join(' ')}} {{(incpaths+incpaths_imported)|prefix('-I')|join(' ')}} -c {inputs} -o {outputs}"
    gnu_compile_pch_cmd = "{gnu_cxx} -x c++-header -MMD -MF {outputs}.d {% if shared %}-fPIC {% endif %}{cxxflags_imported} {cxxflags_environ} {cxxflags} {{(macros_imported+macros)|prefix('-D')|join(' ')}} {{(incpaths+incpaths_imported)|prefix('-I')|join(' ')}} -c {inputs} -o {outputs}"
    gnu_link_cmd = "{gnu_ld} {% if shared %}-shared {% endif %}{ldflags_imported} {ldflags_environ} {ldflags} -Wl,--start-group @{inputs} -Wl,--end-group {{(libpaths+libpaths_imported)|prefix('-L')|join(' ')}} -o {outputs} -Wl,--start-group {{(libraries+libraries_imported)|prefix('-l')|join(' ')}} -Wl,--end-group"

    def gnu_archive(self, binary, objects, **kwargs):
        command = kwargs.pop("command", self.gnu_archive_cmd)

        kwargs["gnu_ar"] = self._gnu_tool("AR", self.gnu_ar)

        outputs = ["{outdir_rel}/lib{binary}.a"]
        objlist = self.gnu_mrilist(objects, outputs[0], "{outdir_rel}/objects.mri")
        command = self.tools.render(command, inputs=objlist, outputs=outputs, **kwargs)

        binary = self.command(
            command,
            inputs=objlist,
            outputs=outputs,
            message="{binary}",
            **kwargs)
        binary.add_dependency(objects)
        return binary

    def gnu_compile_generic(self, cmd, sources=[], **kwargs):
        objects = []

        for srcfile in utils.as_list(sources):
            srcfile = self.tools.expand_relpath(srcfile, self.joltdir)
            objfile = self.tools.expand_relpath(os.path.join(self.outdir, srcfile + ".o"), self.joltdir)
            cmd = self.tools.render(cmd, inputs=srcfile, outputs=objfile, **kwargs)
            obj = self.command(
                cmd,
                srcfile, objfile,
                message="{inputs}",
                **kwargs)
            obj.add_influence_depfile(objfile + ".d")
            if hasattr(self, "_pch_objects"):
                obj.add_dependency(self._pch_objects)
            objects.append(obj)

        return objects

    def gnu_compile_asm(self, sources, **kwargs):
        kwargs["gnu_as"] = self._gnu_tool("AS", self.gnu_as)

        kwargs["asflags"] = kwargs.pop("asflags", self._asflags())
        kwargs["asflags_environ"] = kwargs.pop("asflags_environ", self.tools.getenv("ASFLAGS"))
        kwargs["asflags_imported"] = kwargs.pop("asflags_imported", self._import_listattr("asflags"))
        return self.gnu_compile_generic(self.gnu_compile_asm_cmd, sources, **kwargs)

    def gnu_compile_asm_cpp(self, sources, **kwargs):
        kwargs["gnu_as"] = self._gnu_tool("AS", self.gnu_as)

        kwargs["asflags"] = kwargs.pop("asflags", self._asflags())
        kwargs["asflags_environ"] = kwargs.pop("asflags_environ", self.tools.getenv("ASFLAGS"))
        kwargs["asflags_imported"] = kwargs.pop("asflags_imported", self._import_listattr("asflags"))
        kwargs["incpaths"] = ["{outdir_rel}"] + kwargs.pop("incpaths", self._task_incpaths())
        kwargs["incpaths_imported"] = kwargs.pop("incpaths_imported", self._imported_incpaths())
        kwargs["macros"] = kwargs.pop("macros", self._macros())
        kwargs["macros_imported"] = kwargs.pop("macros_imported", self._import_listattr("macros"))
        return self.gnu_compile_generic(self.gnu_compile_asm_cpp_cmd, sources, **kwargs)

    def gnu_compile_c(self, sources, **kwargs):
        kwargs["gnu_cc"] = self._gnu_tool("CC", self.gnu_cc)

        kwargs["cflags"] = kwargs.pop("cflags", self._cflags())
        kwargs["cflags_environ"] = kwargs.pop("cflags_environ", self.tools.getenv("CFLAGS"))
        kwargs["cflags_imported"] = kwargs.pop("cflags_imported", self._import_listattr("cflags"))
        kwargs["incpaths"] = ["{outdir_rel}"] + kwargs.pop("incpaths", self._task_incpaths())
        kwargs["incpaths_imported"] = kwargs.pop("incpaths_imported", self._imported_incpaths())
        kwargs["macros"] = kwargs.pop("macros", self._macros())
        kwargs["macros_imported"] = kwargs.pop("macros_imported", self._import_listattr("macros"))
        return self.gnu_compile_generic(self.gnu_compile_c_cmd, sources, **kwargs)

    def gnu_compile_cxx(self, sources, **kwargs):
        kwargs["gnu_cxx"] = self._gnu_tool("CXX", self.gnu_cxx)

        kwargs["cxxflags"] = kwargs.pop("cxxflags", self._cxxflags())
        kwargs["cxxflags_environ"] = kwargs.pop("cxxflags_environ", self.tools.getenv("CXXFLAGS"))
        kwargs["cxxflags_imported"] = kwargs.pop("cxxflags_imported", self._import_listattr("cxxflags"))
        kwargs["incpaths"] = ["{outdir_rel}"] + kwargs.pop("incpaths", self._task_incpaths())
        kwargs["incpaths_imported"] = kwargs.pop("incpaths_imported", self._imported_incpaths())
        kwargs["macros"] = kwargs.pop("macros", self._macros())
        kwargs["macros_imported"] = kwargs.pop("macros_imported", self._import_listattr("macros"))

        return self.gnu_compile_generic(self.gnu_compile_cxx_cmd, sources, **kwargs)

    def gnu_compile_h(self, sources, **kwargs):
        return []

    def gnu_compile_pch(self, sources, **kwargs):
        raise_task_error_if(hasattr(self, "_pch_objects"), self, "Multiple precompiled headers found, only one allowed")

        kwargs["gnu_cxx"] = self._gnu_tool("CXX", self.gnu_cxx)
        kwargs["cxxflags"] = kwargs.pop("cxxflags", self._cxxflags())
        kwargs["cxxflags_environ"] = kwargs.pop("cxxflags_environ", self.tools.getenv("CXXFLAGS"))
        kwargs["cxxflags_imported"] = kwargs.pop("cxxflags_imported", self._import_listattr("cxxflags"))
        kwargs["incpaths"] = ["{outdir_rel}"] + kwargs.pop("incpaths", self._task_incpaths())
        kwargs["incpaths_imported"] = kwargs.pop("incpaths_imported", self._imported_incpaths())
        kwargs["macros"] = kwargs.pop("macros", self._macros())
        kwargs["macros_imported"] = kwargs.pop("macros_imported", self._import_listattr("macros"))

        sources = self._to_subtask_list(sources)
        objects = []

        for srcsubtask in sources:
            for srcfile in srcsubtask.outputs:
                srcfile = os.path.relpath(srcfile, self.joltdir)
                objfile = os.path.basename(srcfile)
                objfile = os.path.relpath(os.path.join(self.outdir, objfile + ".gch"), self.joltdir)
                cmd = self.tools.render(self.gnu_compile_pch_cmd, inputs=srcfile, outputs=objfile, **kwargs)
                obj = self.command(
                    cmd,
                    srcfile, objfile,
                    message="{inputs}",
                    **kwargs)
                obj.add_influence_depfile(objfile + ".d")
                objects.append(obj)

        self._pch_objects = objects

        return objects

    def gnu_link(self, binary, objects, **kwargs):
        command = kwargs.pop("command", self.gnu_link_cmd)

        kwargs["gnu_ld"] = self._gnu_tool("CXX", self.gnu_ld)

        kwargs["ldflags"] = kwargs.pop("ldflags", self._ldflags())
        kwargs["ldflags_environ"] = kwargs.pop("ldflags_environ", self.tools.getenv("LDFLAGS"))
        kwargs["ldflags_imported"] = kwargs.pop("ldflags_imported", self._import_listattr("ldflags"))

        kwargs["libpaths"] = kwargs.pop("libpaths", self._libpaths())
        kwargs["libpaths_imported"] = kwargs.pop("libpaths_imported", self._import_pathlistattr("libpaths"))

        kwargs["libraries"] = kwargs.pop("libraries", self._libraries())
        kwargs["libraries_imported"] = kwargs.pop("libraries_imported", self._import_listattr("libraries"))

        objlist = self.filelist(objects, "{outdir}/objects.list")
        if getattr(self, "shared", False):
            outputs = ["{outdir_rel}/lib{binary}.so"]
        else:
            outputs = ["{outdir_rel}/{binary}"]
        command = self.tools.render(command, inputs=objlist, outputs=outputs, **kwargs)

        binary = self.command(
            command,
            inputs=objlist,
            outputs=outputs,
            message="{binary}",
            **kwargs)
        binary.add_dependency(objects)
        return binary

    def gnu_mrilist(self, inputs, target, outputs):
        inputs = self._to_subtask_list(inputs)
        inputs = self._to_output_files(inputs)
        target = self.tools.expand(target)
        mrilist = self.render(
            "create {{target}}\n{% for input in inputs %}{% if input[-2:] == '.a' %}addlib {{input}}\n{% else %}addmod {{ input }}\n{% endif %}{% endfor %}save\nend\n",
            outputs, target=target, inputs=inputs)

        return mrilist

    def gnu_import_libraries(self, deps):
        archives = []
        for name, artifact in deps.items():
            for lib in artifact.cxxinfo.libraries.items():
                name = f"lib{lib}.a"
                for path in artifact.cxxinfo.libpaths.items():
                    archive = os.path.join(artifact.path, path, name)
                    if os.path.exists(archive):
                        sandbox = self.tools.sandbox(artifact, self.incremental)
                        archive = os.path.join(sandbox, path, name)
                        archives.append(archive)
        return archives


def flatbuffer():
    def decorate(cls):
        class WithFlatbufferCompilation(cls):
            abstract = True

            flatc = "flatc"
            flatbuffer_cmd = "{{flatc}} --{{generator}} {fbflags} {{incpaths|prefix('-I ')|join(' ')}} -o {{outdir}} {inputs} && {{flatc}} -M --{{generator}} {fbflags} {{incpaths|prefix('-I ')|join(' ')}} -o {{outdir}} {inputs} > {depfile}"
            fbflags = []

            def run(self, deps, tools):
                super().run(deps, tools)

            def compile_flatbuffer(self, inputs, **kwargs):
                kwargs["flatc"] = self._gnu_tool("FLATC", self.flatc)

                jobs = []

                for input in inputs:
                    context = self.context(input, **kwargs)
                    context["generator"] = kwargs.pop("generator", "cpp")

                    command = self.tools.render(self.flatbuffer_cmd, **context)
                    depfile = self.tools.expand("{outdir}/{input}.d", input=input, **context)
                    outputs = []

                    depfiledir = self.mkdirname(depfile)

                    flatbuffer_job = self.command(
                        command,
                        inputs=input,
                        outputs=self.tools.expand(outputs, **context),
                        message="{inputs} [{generator}]",
                        depfile=depfile,
                        **context)

                    flatbuffer_job.add_dependency(depfiledir)
                    flatbuffer_job.add_influence_depfile(depfile)
                    jobs.append(flatbuffer_job)

                return jobs

            def compile(self, sources, **kwargs):
                flatbuffers, sources = self.filter_by_ext(sources, ".fbs")
                if not flatbuffers:
                    return super().compile(sources, **kwargs)

                flatbuffer_job = self.compile_flatbuffer(flatbuffers, **kwargs)
                flatbuffer_sources = self._to_output_files(flatbuffer_job)
                flatbuffer_outputs = self.compile(flatbuffer_sources, **kwargs)

                outputs = self.compile(sources, **kwargs)
                for output in outputs:
                    output.add_dependency(flatbuffer_outputs)

                return outputs + flatbuffer_outputs

        return WithFlatbufferCompilation

    return decorate


def protobuf():
    def decorate(cls):
        class WithProtobufCompilation(cls):
            abstract = True

            protoc = "protoc"
            proto_cmd = "{protoc} {{(incpaths_imported+incpaths)|prefix('-I')|join(' ')}} {protoflags} --dependency_out={depfile} --{generator}_out={outdir_rel}{% if generator_plugin %} --plugin=protoc-gen-{generator}={generator_plugin}{% endif %} {inputs}"
            protoflags = []
            protogenerator = "cpp"

            def run(self, deps, tools):
                self.protoc = self._gnu_tool("PROTOC", self.protoc)
                super().run(deps, tools)

            def context(self, input=None, **kwargs):
                kwargs = copy(kwargs)

                kwargs["asflags"] = kwargs.pop("cflags", self._cflags())
                kwargs["asflags_imported"] = kwargs.pop("asflags_imported", self._import_listattr("asflags"))

                kwargs["cflags"] = kwargs.pop("cflags", self._cflags())
                kwargs["cflags_imported"] = kwargs.pop("cflags_imported", self._import_listattr("cflags"))

                kwargs["cxxflags"] = kwargs.pop("cxxflags", self._cxxflags())
                kwargs["cxxflags_imported"] = kwargs.pop("cxxflags_imported", self._import_listattr("cxxflags"))

                kwargs["incpaths"] = ["{outdir_rel}"] + kwargs.pop("incpaths", self._task_incpaths())
                kwargs["incpaths_imported"] = kwargs.pop("incpaths_imported", self._imported_incpaths())

                if hasattr(self, "_ldflags"):
                    kwargs["ldflags"] = kwargs.pop("ldflags", self._ldflags())
                    kwargs["ldflags_imported"] = kwargs.pop("ldflags_imported", self._import_listattr("ldflags"))

                if hasattr(self, "_libpaths"):
                    kwargs["libpaths"] = kwargs.pop("libpaths", self._libpaths())
                    kwargs["libpaths_imported"] = kwargs.pop("libpaths_imported", self._import_pathlistattr("libpaths"))

                if hasattr(self, "_libraries"):
                    kwargs["libraries"] = kwargs.pop("libraries", self._libraries())
                    kwargs["libraries_imported"] = kwargs.pop("libraries_imported", self._import_listattr("libraries"))

                kwargs["macros"] = kwargs.pop("macros", self._macros())
                kwargs["macros_imported"] = kwargs.pop("macros_imported", self._import_listattr("macros"))

                if input:
                    kwargs["in_path"] = os.path.dirname(input) or "."
                    kwargs["in_base"], kwargs["in_ext"] = os.path.splitext(os.path.basename(input))

                return kwargs

            def compile_proto_cpp(self, inputs, **kwargs):
                depfile = "{outdir_rel}/{in_base}.pb.d"
                outputs = [
                    "{outdir_rel}/{in_base}.pb.h",
                    "{outdir_rel}/{in_base}.pb.cc",
                ]
                kwargs["depfile"] = kwargs.pop("depfile", depfile)
                kwargs["generator"] = kwargs.pop("generator", "cpp")
                kwargs["outputs"] = kwargs.pop("outputs", outputs)

                return self._compile_proto(inputs, **kwargs)

            def compile_proto_cpp_grpc(self, inputs, **kwargs):
                return self.compile_proto_cpp(inputs, **kwargs) + \
                    self.compile_proto_grpc(inputs, generator_plugin="grpc_cpp_plugin", **kwargs)

            def compile_proto_grpc(self, inputs, **kwargs):
                depfile = "{outdir_rel}/{in_base}.grpc.pb.d"
                outputs = [
                    "{outdir_rel}/{in_base}.grpc.pb.h",
                    "{outdir_rel}/{in_base}.grpc.pb.cc",
                ]
                kwargs["depfile"] = kwargs.pop("depfile", depfile)
                kwargs["generator"] = kwargs.pop("generator", "grpc")
                kwargs["generator_plugin"] = self.tools.which(kwargs.pop("generator_plugin", "grpc_cpp_plugin"))
                kwargs["outputs"] = kwargs.pop("outputs", outputs)
                return self._compile_proto(inputs, **kwargs)

            def compile_proto_python(self, inputs, **kwargs):
                depfile = "{outdir_rel}/{in_base}_pb2.py.d"
                outputs = [
                    "{outdir_rel}/{in_base}_pb2.py",
                ]
                kwargs["command"] = kwargs.pop("command", self.proto_cmd)
                kwargs["depfile"] = kwargs.pop("depfile", depfile)
                kwargs["generator"] = kwargs.pop("generator", "python")
                kwargs["outputs"] = kwargs.pop("outputs", outputs)
                return self._compile_proto(inputs, **kwargs)

            def compile_proto_python_grpc(self, inputs, **kwargs):
                depfile = "{outdir_rel}/{in_base}_pb2_grpc.py.d"
                outputs = [
                    "{outdir_rel}/{in_base}_pb2_grpc.py",
                ]
                kwargs["protoc"] = kwargs.pop("protoc", "python -m grpc_tools.protoc")
                kwargs["command"] = kwargs.pop("command", self.proto_cmd)
                kwargs["depfile"] = kwargs.pop("depfile", depfile)
                kwargs["generator"] = kwargs.pop("generator", "grpc_python")
                kwargs["outputs"] = kwargs.pop("outputs", outputs)
                return self._compile_proto(inputs, **kwargs)

            def _compile_proto(self, inputs, **kwargs):
                jobs = []

                for input in inputs:
                    context = self.context(input, **kwargs)

                    command = self.tools.render(context.pop("command", self.proto_cmd), **context)
                    depfile = context.pop("depfile")
                    depfile = self.tools.expand(depfile, **context)
                    outputs = context.pop("outputs")

                    proto_job = self.command(
                        command,
                        inputs=input,
                        outputs=self.tools.expand(outputs, **context),
                        message="{inputs} [{generator}]",
                        depfile=depfile,
                        **context)

                    proto_job.add_influence_depfile(depfile)
                    jobs.append(proto_job)

                return jobs

            def compile(self, sources, **kwargs):
                protobufs, sources = self.filter_by_ext(sources, ".proto")
                if not protobufs:
                    return super().compile(sources, **kwargs)

                proto_sources = []
                generators = utils.as_list(kwargs.get("protogenerator", getattr(self, "protogenerator", [])))
                for generator in generators:
                    proto_fn = getattr(self, "compile_proto_" + generator, None)
                    raise_task_error_if(not proto_fn, self, "Protobuf generator '{}' is not supported", generator)
                    proto_job = proto_fn(protobufs, **kwargs)
                    proto_sources += self._to_output_files(proto_job)
                proto_outputs = self.compile(proto_sources, **kwargs)

                outputs = self.compile(sources, **kwargs)
                for output in outputs:
                    output.add_dependency(proto_job)

                return outputs + proto_outputs

        WithProtobufCompilation.__doc__ = cls.__doc__
        WithProtobufCompilation.__name__ = cls.__name__
        return WithProtobufCompilation

    return decorate


@attributes.method("archive", "{toolchain}_archive")
@attributes.method("compile_asm", "{toolchain}_compile_asm")
@attributes.method("compile_asm_cpp", "{toolchain}_compile_asm_cpp")
@attributes.method("compile_c", "{toolchain}_compile_c")
@attributes.method("compile_cxx", "{toolchain}_compile_cxx")
@attributes.method("compile_h", "{toolchain}_compile_h")
@attributes.method("compile_pch", "{toolchain}_compile_pch")
@attributes.method("import_libraries", "{toolchain}_import_libraries")
@attributes.method("link", "{toolchain}_link")
@ninja.attributes.asflags("asflags")
@ninja.attributes.cflags("cflags")
@ninja.attributes.cxxflags("cxxflags")
@ninja.attributes.incpaths("incpaths")
@ninja.attributes.ldflags("ldflags")
@ninja.attributes.libpaths("libpaths")
@ninja.attributes.libraries("libraries")
@ninja.attributes.macros("macros")
@ninja.attributes.sources("sources")
@flatbuffer()
@protobuf()
class Compilation(GnuToolchain, MultiTask):
    abstract = True

    binary = "{canonical_name}"

    incremental = True

    source_influence = True
    """ Let the contents of source files influence the identity of the task artifact.

    When ``True``, a source file listed in the ``sources`` attribute will
    cause a rebuild of the task if modified.

    Source influence can hurt performance since every files needs to be hashed.
    It is safe to set this flag to ``False`` if all source files reside in a
    ``git`` repository listed as a dependency with the ``requires`` attribute or
    if the task uses the ``git.influence`` decorator.

    Always use ``source_influence`` if you are unsure whether it is needed or not.
    """

    toolchain = "gnu"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.binary = self.expand(utils.call_or_return(self, self.__class__.binary))
        if self.source_influence:
            for source in self._sources():
                self.influence.append(FileInfluence(source))

    def _import_listattr(self, attribute):
        result = []
        for _, artifact in self.deps.items():
            result.extend(list(getattr(artifact.cxxinfo, attribute, []).items()))
        return result

    def _import_pathlistattr(self, attribute):
        result = []
        for _, artifact in self.deps.items():
            result.extend([os.path.join(artifact.path, path) for path in list(getattr(artifact.cxxinfo, attribute, []).items())])
        return result

    @utils.cached.instance
    def _task_incpaths(self):
        result = []

        for path in self._incpaths():
            if path[0] in ['=', os.sep]:
                path = self.tools.expand(path)
            if path[0] in ['-']:
                path = self.tools.expand_relpath(path[1:], self.joltdir)
            result.append(path)

        return result

    @utils.cached.instance
    def _imported_incpaths(self):
        self.depsdir = self.tools.builddir("deps", incremental=self.incremental)

        result = []
        for _, artifact in self.deps.items():
            symlink = os.path.join(self.depsdir, utils.canonical(artifact.task.short_qualified_name), artifact.name)
            self.tools.symlink(artifact.path, symlink, replace=True)

            for path in list(getattr(artifact.cxxinfo, "incpaths", []).items()):
                if path[0] in ['=', os.sep]:
                    path = self.tools.expand(path)
                    result.append(path)
                elif path[0] in ['-']:
                    path = self.tools.expand_relpath(path[1:], self.joltdir)
                    result.append(path)
                else:
                    path = os.path.join(symlink, path)
                    path = os.path.relpath(path, self.joltdir)
                    result.append(path)
        return result

    def filelist(self, inputs, outputs):
        inputs = self._to_subtask_list(inputs)
        inputs = self._to_output_files(inputs)
        objlist = self.render(
            "{% for input in inputs %}{{ input }}\n{% endfor %}",
            outputs, inputs=inputs)

        return objlist

    def filter_by_ext(self, inputs, exts):
        exts = utils.as_list(exts)
        inputs = self._to_subtask_list(inputs)
        inputs = self._to_output_files(inputs)

        matches = []
        mismatches = []
        for input in inputs:
            _, ext = os.path.splitext(input)
            if ext in exts:
                matches.append(input)
            else:
                mismatches.append(input)

        return matches, mismatches

    def glob(self, paths):
        result = []
        for path in paths:
            found = self.tools.glob(path)
            raise_task_error_if(
                not found and not ('*' in path or '?' in path), self,
                "Listed source file '{0}' not found in workspace", os.path.basename(path))
            result += found
        return result

    def glob_headers(self):
        return self.glob(self._headers())

    def glob_sources(self, deps=None):
        sources_imported = self._import_pathlistattr("sources") if deps else []
        return self.glob(self._sources() + sources_imported)

    def compile(self, sources):
        outputs = []
        newsources = []

        for source in utils.as_list(sources):
            _, ext = os.path.splitext(source)
            if ext in [".pch"]:
                self.compile_pch(source)
            else:
                pass

        sources += newsources

        for source in utils.as_list(sources):
            _, ext = os.path.splitext(source)
            if ext in [".a"]:
                outputs.extend(self.compile_lib(source))
            elif ext in [".c"]:
                outputs.extend(self.compile_c(source))
            elif ext in [".cc", ".cpp", ".cxx"]:
                outputs.extend(self.compile_cxx(source))
            elif ext in [".h"]:
                outputs.extend(self.compile_h(source))
            elif ext in [".pch"]:
                pass
            elif ext in [".py"]:
                pass
            elif ext in [".s"]:
                outputs.extend(self.compile_asm(source))
            elif ext in [".S"]:
                outputs.extend(self.compile_asm_cpp(source))
            else:
                raise_task_error(self, "Unknown file extension: {}", source)

        return outputs

    def compile_lib(self, sources):
        return self._to_subtask_list(sources)

    def generate(self, deps, tools):
        self.deps = deps
        self.outdir = self.tools.builddir("outputs", incremental=True)
        self.outdir_rel = self.tools.expand_relpath(self.outdir, self.joltdir)

        sources = self.glob_sources(deps)
        return self.compile(sources)


@ninja.attributes.headers("headers")
class Library(Compilation):
    abstract = True

    headers = []

    publishapi = "include/"
    """ The artifact path where the library is published. """

    publishdir = "lib/"
    """ The artifact path where the library is published. """

    shared = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.source_influence:
            for header in self._headers():
                self.influence.append(FileInfluence(header))

    def generate(self, deps, tools):
        self.deps = deps
        self.outdir = self.tools.builddir("outputs", incremental=True)
        self.outdir_rel = self.tools.expand_relpath(self.outdir, self.joltdir)

        sources = self.glob_sources(deps)
        if self.selfsustained:
            sources += self.import_libraries(deps)
        objects = self.compile(sources)
        if self.shared:
            binary = self.link(self.binary, objects)
        else:
            binary = self.archive(self.binary, objects)
        return binary

    def publish(self, artifact, tools):
        for header in self._headers():
            artifact.collect(header, self.publishapi)

        with tools.cwd(self.outdir):
            if not self.shared:
                if artifact.collect("lib{binary}.a", self.publishdir):
                    artifact.cxxinfo.libpaths.append(self.publishdir)
                    artifact.cxxinfo.libraries.append("{binary}")
            else:
                if artifact.collect("lib{binary}.so", self.publishdir):
                    artifact.cxxinfo.libpaths.append(self.publishdir)
                    artifact.cxxinfo.libraries.append("{binary}")
                artifact.collect("{binary}.dll", self.publishdir)
            if artifact.collect("{binary}.lib", self.publishdir):
                artifact.cxxinfo.libpaths.append(self.publishdir)
                artifact.cxxinfo.libraries.append("{binary}")


class Executable(Compilation):
    abstract = True

    publishdir = "bin/"
    """ The artifact path where the binary is published. """

    strip = True
    """
    Remove debug information from binary.

    When using the GNU toolchain, debug information is kept in a separate binary
    which is either published or not depending on the value of this attribute.
    It's found in a .debug directory if present.

    Only applicable to shared libraries.
    """

    def generate(self, deps, tools):
        self.deps = deps
        self.outdir = self.tools.builddir("outputs", incremental=True)
        self.outdir_rel = self.tools.expand_relpath(self.outdir, self.joltdir)

        sources = self.glob_sources(deps)
        objects = self.compile(sources)
        binary = self.link(self.binary, objects)
        return binary

    def publish(self, artifact, tools):
        with tools.cwd(self.outdir):
            if os.name == "nt":
                binary = artifact.collect(self.binary + '.exe', self.publishdir)
            else:
                binary = artifact.collect(self.binary, self.publishdir)
            if not self.strip:
                artifact.collect(".debug", self.publishdir)
            if binary:
                artifact.environ.PATH.append(self.publishdir)
                artifact.strings.executable = os.path.join(
                    self.publishdir, self.binary)



from jolt import Parameter, Export
from jolt import log
from jolt.tasks import TaskRegistry
from jolt.plugins import git


log.verbose("[Gerrit] Loaded")


class GerritSrc(git.GitSrc):
    name = "gerrit-src"
    url = Parameter(help="URL to the Gerrit git repo to be cloned. Required.")
    sha = Parameter(required=False, help="Specific commit or tag to be checked out. Optional.")
    path = Parameter(required=False, help="Local path where the repository should be cloned.")
    _revision = Export(value=lambda self: self._get_revision() or self.git.head())

    def __init__(self, *args, **kwargs):
        refspec1 = '+refs/changes/*:refs/remotes/origin/changes/*'
        super().__init__(*args, refspecs=[refspec1], **kwargs)


class Gerrit(git.Git):
    name = "gerrit"
    url = Parameter(help="URL to the Gerrit git repo to be cloned. Required.")
    sha = Parameter(required=False, help="Specific commit or tag to be checked out. Optional.")
    path = Parameter(required=False, help="Local path where the repository should be cloned.")
    _revision = Export(value=lambda self: self._get_revision() or self.git.head())

    def __init__(self, *args, **kwargs):
        refspec1 = '+refs/changes/*:refs/remotes/origin/changes/*'
        super().__init__(*args, refspecs=[refspec1], **kwargs)


TaskRegistry.get().add_task_class(GerritSrc)
TaskRegistry.get().add_task_class(Gerrit)



from copy import copy
from threading import RLock
from xml.dom import minidom

from jolt.influence import HashInfluenceProvider, HashInfluenceRegistry
from jolt.tools import Tools
from jolt import filesystem as fs
from jolt import log
from jolt.plugins import git
from jolt.scheduler import NetworkExecutorExtension, NetworkExecutorExtensionFactory
from jolt.xmldom import Attribute, Composition, SubElement, Element, ElementTree, ET


@Attribute('name')
@Attribute('fetch')
class RepoRemote(SubElement):
    def __init__(self, elem=None):
        super(RepoRemote, self).__init__('remote', elem=elem)


@Attribute('revision')
@Attribute('remote')
@Attribute('sync-j')
class RepoDefault(SubElement):
    def __init__(self, elem=None):
        super(RepoDefault, self).__init__('default', elem=elem)


@Attribute('path')
@Attribute('name')
@Attribute('revision')
@Attribute('upstream')
@Attribute('remote')
class RepoProject(SubElement):
    def __init__(self, elem=None):
        super(RepoProject, self).__init__('project', elem=elem)

    def get_diff(self):
        with self.tools.cwd(self.path_or_name):
            return self.tools.run("git diff --no-ext-diff HEAD", output_on_error=True)
        assert False, "git command failed"

    def get_head(self):
        with self.tools.cwd(self.path_or_name):
            return self.tools.run("git rev-parse HEAD", output_on_error=True).strip()
        assert False, "git command failed"

    def get_remote_branches(self, commit):
        with self.tools.cwd(self.path_or_name):
            result = self.tools.run("git branch -r --contains {0}", commit, output_on_error=True)
            if not result:
                return []
            result = result.strip().splitlines()
            result = [line.strip() for line in result]
            return result
        assert False, "git command failed"

    def get_local_commits(self):
        with self.tools.cwd(self.path_or_name):
            result = self.tools.run("git rev-list HEAD ^@{{upstream}}", output_on_error=True)
            if not result:
                return []
            result = result.strip("\r\n ")
            return result.splitlines()
        assert False, "git command failed"

    def has_local_ref(self, ref):
        with self.tools.cwd(self.path_or_name):
            try:
                return True if self.tools.run("git show-ref {0}", ref, output=False) else False
            except Exception:
                return False

    def get_remote_ref(self, commit, remote, pattern=None):
        with self.tools.cwd(self.path_or_name):
            result = self.tools.run(
                "git ls-remote {0}{1}",
                remote,
                " {0}".format(pattern) if pattern else "",
                output_on_error=True)
            if not result:
                return None
            result = result.strip().splitlines()
            result = [line.split("\t") for line in result]
            result = {line[0]: line[1] for line in result}
            return result.get(commit)
        assert False, "git command failed"

    @property
    def path_or_name(self):
        return self.path or self.name


@Composition(RepoRemote, "remote")
@Composition(RepoDefault, "default")
@Composition(RepoProject, "project")
class RepoManifest(ElementTree):
    def __init__(self, task, path):
        super(RepoManifest, self).__init__(element=Element('manifest'))
        self.path = path
        self.tools = task.tools

    def append(self, element):
        self.getroot().append(element)

    def get(self, key):
        self.getroot().get(key)

    def set(self, key, value):
        self.getroot().set(key, value)

    def parse(self, filename=".repo/manifest.xml"):
        with open(fs.path.join(self.tools.getcwd(), filename)) as f:
            root = ET.fromstring(f.read())
            self._setroot(root)
            for project in self.projects:
                project.tools = self.tools
            return self
        raise Exception("failed to parse xml file")

    def format(self):
        return minidom.parseString(ET.tostring(self.getroot())).toprettyxml(indent="  ")

    def write(self, filename):
        with open(filename, 'w') as f:
            f.write(self.format())

    def get_remote(self, project):
        if len(self.defaults) > 0:
            return project.remote or self.defaults[0].remote
        return project.remote

    def get_upstream(self, project):
        if len(self.defaults) > 0:
            return project.upstream or self.defaults[0].revision or "master"
        return project.upstream or "master"

    def assert_clean(self):
        for project in self.projects:
            assert not project.get_diff(), \
                "repo project '{0}' has local changes"\
                .format(project.path_or_name)

    def lock_revisions(self):
        for project in self.projects:
            head = project.get_head()
            if not project.get_remote_branches(head):
                remote_ref = project.get_remote_ref(
                    head, self.get_remote(project))
                assert remote_ref, \
                    "repo project '{0}' has unpublished commits"\
                    .format(project.path_or_name)
                head = remote_ref

            if not project.upstream:
                if project.revision and project.has_local_ref(project.revision):
                    project.upstream = project.revision
                else:
                    project.upstream = self.get_upstream(project)

            if not project.revision.startswith("refs/changes"):
                project.revision = head


_git_repos = {}
_git_repo_lock = RLock()

_repo_manifests = {}
_repo_manifest_lock = RLock()


class RepoInfluenceProvider(HashInfluenceProvider):
    name = "Repo"
    path = "."

    def __init__(self, path=None, include=None, exclude=None, network=True):
        self.path = path or self.__class__.path
        self.include = include
        self.exclude = exclude
        self.network = network

    def get_influence(self, task):
        self.tools = Tools(task, task.joltdir)
        try:
            manifest_path = fs.path.join(task.joltdir, task.expand(self.path))
            manifest = RepoManifest(task, manifest_path)
            manifest.parse(fs.path.join(manifest_path, ".repo", "manifest.xml"))

            result = []
            for project in sorted(manifest.projects, key=lambda p: p.name):
                if self.include is not None and project.path_or_name not in self.include:
                    continue
                if self.exclude is not None and project.path_or_name in self.exclude:
                    continue

                with _git_repo_lock:
                    gip = _git_repos.get(project.path_or_name)
                    if gip is None:
                        gip = git.GitInfluenceProvider(project.path_or_name)
                        _git_repos[project.path_or_name] = gip
                    result.append(gip.get_influence(task))

            return "\n".join(result)

        except KeyError:
            log.exception()
        assert False, "failed to calculate hash influence for repo manifest at {0}".format(self.path)

    def get_manifest(self, task):
        manifest_path = fs.path.join(task.joltdir, task.expand(self.path))
        with _repo_manifest_lock:
            manifest = _repo_manifests.get(manifest_path)
            if manifest is None:
                manifest = RepoManifest(task, manifest_path)
                manifest.parse()
                manifest.assert_clean()
                manifest.lock_revisions()
                _repo_manifests[manifest_path] = manifest
            return manifest


class RepoNetworkExecutorExtension(NetworkExecutorExtension):
    def get_parameters(self, task):
        rip = list(filter(
            lambda n: isinstance(n, RepoInfluenceProvider),
            task.task.influence))
        if rip:
            assert len(rip) == 1, "task influenced by multiple repo manifests"
            rip = rip[0]
            if rip.network:
                manifest = rip.get_manifest(task.task)
                return {"repo_manifest": manifest.format()}
        return {}


@NetworkExecutorExtensionFactory.Register
class RepoNetworkExecutorExtensionFactory(NetworkExecutorExtensionFactory):
    def create(self):
        return RepoNetworkExecutorExtension()


def global_influence(path, include=None, exclude=None, network=True, cls=RepoInfluenceProvider):
    HashInfluenceRegistry.get().register(cls(path, include, exclude, network))


def influence(path='.', include=None, exclude=None, network=True, cls=RepoInfluenceProvider):
    def _decorate(taskcls):
        if "influence" not in taskcls.__dict__:
            taskcls.influence = copy(taskcls.influence)
        provider = cls(path, include, exclude, network)
        taskcls.influence.append(provider)
        return taskcls
    return _decorate



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from jolt.plugins.remote_execution import scheduler_pb2 as jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2
from jolt.plugins.remote_execution import worker_pb2 as jolt_dot_plugins_dot_remote__execution_dot_worker__pb2


class WorkerStub(object):
    """The scheduler interface for workers and executors.
    Workers prepare the build environment by deploying and
    running executors (Jolt clients) to execute tasks.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetInstructions = channel.stream_stream(
                '/Worker/GetInstructions',
                request_serializer=jolt_dot_plugins_dot_remote__execution_dot_worker__pb2.WorkerUpdate.SerializeToString,
                response_deserializer=jolt_dot_plugins_dot_remote__execution_dot_worker__pb2.WorkerRequest.FromString,
                )
        self.GetTasks = channel.stream_stream(
                '/Worker/GetTasks',
                request_serializer=jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2.TaskUpdate.SerializeToString,
                response_deserializer=jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2.TaskRequest.FromString,
                )


class WorkerServicer(object):
    """The scheduler interface for workers and executors.
    Workers prepare the build environment by deploying and
    running executors (Jolt clients) to execute tasks.
    """

    def GetInstructions(self, request_iterator, context):
        """Called by a worker to get build instructions.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTasks(self, request_iterator, context):
        """Called by an executor (Jolt client) to get task execution requests.
        Task updates, including status and logs are sent back to the scheduler.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_WorkerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetInstructions': grpc.stream_stream_rpc_method_handler(
                    servicer.GetInstructions,
                    request_deserializer=jolt_dot_plugins_dot_remote__execution_dot_worker__pb2.WorkerUpdate.FromString,
                    response_serializer=jolt_dot_plugins_dot_remote__execution_dot_worker__pb2.WorkerRequest.SerializeToString,
            ),
            'GetTasks': grpc.stream_stream_rpc_method_handler(
                    servicer.GetTasks,
                    request_deserializer=jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2.TaskUpdate.FromString,
                    response_serializer=jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2.TaskRequest.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Worker', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Worker(object):
    """The scheduler interface for workers and executors.
    Workers prepare the build environment by deploying and
    running executors (Jolt clients) to execute tasks.
    """

    @staticmethod
    def GetInstructions(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/Worker/GetInstructions',
            jolt_dot_plugins_dot_remote__execution_dot_worker__pb2.WorkerUpdate.SerializeToString,
            jolt_dot_plugins_dot_remote__execution_dot_worker__pb2.WorkerRequest.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTasks(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/Worker/GetTasks',
            jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2.TaskUpdate.SerializeToString,
            jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2.TaskRequest.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



"""
Helper module that sets up a mount and user namespace before executing
a command. The helper primarily bypasses the restriction on creating
namespaces from multithreaded applications (which Jolt is).
"""
import argparse
from ctypes import CDLL, c_char_p
import multiprocessing
import os
import shutil
import sys

libc = CDLL("libc.so.6")

CLONE_NEWNS = 0x00020000
CLONE_NEWUSER = 0x10000000

MS_RDONLY = 1
MS_BIND = 4096
MS_REC = 16384


def mount_bind(src, dst, ro=False):
    src = os.path.normpath(src)
    dst = os.path.normpath(dst)

    if os.path.isdir(src):
        os.makedirs(dst, exist_ok=True)
    else:
        os.makedirs(os.path.dirname(dst), exist_ok=True)
        if not os.path.exists(dst):
            with open(dst, "a"):
                pass

    assert libc.mount(
        c_char_p(src.encode("utf-8")),
        c_char_p(dst.encode("utf-8")),
        None,
        MS_BIND | MS_REC | (MS_RDONLY if ro else 0),
        None) == 0


def mount_overlay(src, dst, temp):
    src = os.path.normpath(src)
    dst = os.path.normpath(dst)

    upper = os.path.join(temp, "upper")
    work = os.path.join(temp, "work")
    os.makedirs(upper, exist_ok=True)
    os.makedirs(work, exist_ok=True)

    return libc.mount(
        c_char_p("none".encode("utf-8")),
        c_char_p(dst.encode("utf-8")),
        c_char_p("overlay".encode("utf-8")),
        0,
        f"lowerdir={src},upperdir={upper},workdir={work}".encode("utf-8")) == 0


def mount_tmpfs(path):
    assert libc.mount(
        c_char_p("none".encode("utf-8")),
        c_char_p(path.encode("utf-8")),
        c_char_p("tmpfs".encode("utf-8")),
        0, None) == 0, f"Failed to mount tmpfs at '{path}'"


def main():
    parser = argparse.ArgumentParser(
        prog='chroot',
        description='Runs a command in a chroot using linux namespaces')
    parser.add_argument('command', nargs='+')
    parser.add_argument('-b', '--bind', nargs="*")
    parser.add_argument('-c', '--chroot', required=True)
    parser.add_argument('-d', '--chdir')
    parser.add_argument('--shell', default="True")
    args = parser.parse_args()

    cwd = os.getcwd()
    gid = os.getegid()
    uid = os.geteuid()
    gidmap = [gid, gid, 1]
    uidmap = [uid, uid, 1]
    gidmap = [str(i) for i in gidmap]
    uidmap = [str(i) for i in uidmap]

    newgidmap = shutil.which("newgidmap") or shutil.which("/usr/bin/newgidmap")
    newuidmap = shutil.which("newuidmap") or shutil.which("/usr/bin/newuidmap")

    sem = multiprocessing.Semaphore(0)
    parent = os.getpid()
    child = os.fork()
    if child == 0:
        sem.acquire()
        pid = os.fork()
        if pid == 0:
            os.execve(newuidmap, [newuidmap, str(parent)] + uidmap, {})
            os._exit(1)
        _, status = os.waitpid(pid, 0)
        assert status == 0, f"Failed to map UIDs: newuidmap exit status: {status}"
        os.execve(newgidmap, [newgidmap, str(parent)] + gidmap, {})
        os._exit(1)

    assert libc.unshare(CLONE_NEWNS | CLONE_NEWUSER) == 0
    sem.release()
    _, status = os.waitpid(child, 0)
    assert status == 0, f"Failed to map GIDs: newgidmap exit status: {status}"

    mount_tmpfs("/mnt")
    if not mount_overlay(args.chroot, "/mnt", "/mnt"):
        mount_bind(args.chroot, "/mnt", True)
    mount_bind("/proc", "/mnt/proc", True)

    for path in args.bind or []:
        mount_bind(path, os.path.join("/mnt", os.path.relpath(path, "/")))

    os.chroot("/mnt")
    os.chdir(args.chdir or cwd)

    # Adjust PATH to include standard paths, if missing
    path = os.environ.get("PATH", None)
    if path is None:
        path = "/usr/local/bin:/usr/bin:/bin"
    else:
        path = path.split(os.pathsep)
        if "/usr/local/bin" not in path:
            path.append("/usr/local/bin")
        if "/usr/bin" not in path:
            path.append("/usr/bin")
        if "/bin" not in path:
            path.append("/bin")
        path = os.pathsep.join(path)
    os.environ["PATH"] = path

    if args.shell in ["True", "true"]:
        os.execve("/bin/sh", ["sh", "-c", " ".join(args.command)], os.environ)
    else:
        exe = shutil.which(args.command[0])
        os.execve(exe, args.command, os.environ)


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print(f"{type(e).__name__}: {e}", file=sys.stderr)
        os._exit(1)



import pkg_resources

from jolt import config
from jolt import filesystem as fs
from jolt import influence
from jolt import log
from jolt import common_pb2 as common_pb
from jolt import utils
from jolt import version
from jolt.cache import ArtifactCache
from jolt.error import raise_error_if
from jolt.graph import GraphBuilder
from jolt.loader import JoltLoader
from jolt.manifest import JoltManifest
from jolt.scheduler import JoltEnvironment
from jolt.scheduler import LocalExecutor
from jolt.scheduler import LocalExecutorFactory
from jolt.scheduler import NetworkExecutorExtension
from jolt.scheduler import NetworkExecutorExtensionFactory
from jolt.tasks import Task, TaskRegistry


log.verbose("[SelfDeploy] Loaded")

_path = fs.path.dirname(__file__)
_path = fs.path.dirname(_path)
_path = fs.path.dirname(_path)


@influence.files(fs.path.join(_path, "**", "*.py"))
@influence.files(fs.path.join(_path, "**", "*.sh"))
@influence.files(fs.path.join(_path, "**", "*.xslt"))
@influence.files(fs.path.join(_path, "**", "*.template"))
@influence.attribute("dependencies")
@influence.attribute("extra_dependencies")
class Jolt(Task):
    name = "jolt"

    def __init__(self, *args, **kwargs):
        super(Jolt, self).__init__(*args, **kwargs)
        self.influence.append(
            influence.StringInfluence(
                config.get("selfdeploy", "requires", "")))
        for e in self.extra_files:
            self.influence.append(influence.DirectoryInfluence(e))

    def _verify_influence(self, *args, **kwargs):
        pass  # FIXME: Validation doesn't work properly for directories

    @property
    def loaderdir(self):
        return JoltLoader.get().joltdir

    @property
    def extra_files(self):
        ext = config.get("selfdeploy", "extra", "")
        return [fs.path.join(self.loaderdir, e) for e in ext.split(",")] if ext else []

    @property
    def extra_dependencies(self):
        return get_extra_dependencies()

    @property
    def dependencies(self):
        """
        Queries local pip database for packages and versions used by jolt.

        The data is used to pin dependencies at specific versions so that
        the virtalenv installed on a worker matches what the client had
        installed locally.

        If pip is not available or its internal implementation has changed,
        no version pinning will be performed. Instead, workers will install
        Jolt with its default loose version requirements.
        """
        return get_dependencies(["jolt"] + self.extra_dependencies)

    def publish(self, artifact, tools):
        with tools.cwd(tools.builddir()):
            pinned_reqs = self.dependencies
            if pinned_reqs:
                tools.write_file("requirements.txt", "\n".join(pinned_reqs))
                artifact.collect("requirements.txt")
        with tools.cwd(_path):
            artifact.collect('README.rst')
            artifact.collect('setup.py')
            artifact.collect('jolt/*.py')
            artifact.collect('jolt/*.sh')
            artifact.collect('jolt/*/*.py')
            artifact.collect('jolt/*/*/*.py')
            artifact.collect('jolt/*/*.xslt')
            artifact.collect('jolt/*/*.template')
            artifact.collect('jolt/plugins/selfdeploy/README.rst', flatten=True)
            artifact.collect('jolt/plugins/selfdeploy/setup.py', flatten=True)
            for e in self.extra_files:
                with tools.cwd(fs.path.dirname(e)):
                    artifact.collect(fs.path.basename(e))


class SelfDeployExtension(NetworkExecutorExtension):
    @utils.cached.instance
    def get_parameters(self, _):
        client = get_client()
        params = {}
        if client.identity:
            params["jolt_identity"] = client.identity
        if client.url:
            params["jolt_url"] = client.url
        return params


@NetworkExecutorExtensionFactory.Register
class SelfDeployExtensionFactory(NetworkExecutorExtensionFactory):
    def create(self):
        return SelfDeployExtension()


@utils.cached.method
def get_dependencies(packages=None):
    reqs = packages or ["jolt"]
    pkgs = {}

    while reqs:
        req = reqs.pop()

        dist = pkg_resources.working_set.by_key.get(req)
        if dist is None:
            log.debug("[SelfDeploy] Dependency not found: {}", req)
            pkgs[req] = req
            continue

        for dep in dist.requires():
            if dep.name not in pkgs:
                reqs.append(dep.name)

        pkgs[req] = f"{dist.project_name}=={dist.version}"

    try:
        del pkgs["jolt"]
    except KeyError:
        pass

    return list(sorted(pkgs.values()))


@utils.cached.method
def get_extra_dependencies():
    req = config.get("selfdeploy", "requires", "")
    return req.split() if req else []


@utils.cached.method
def publish_artifact():
    registry = TaskRegistry()
    registry.add_task_class(Jolt)
    acache = ArtifactCache.get()
    env = JoltEnvironment(cache=acache)
    gb = GraphBuilder(registry, acache, JoltManifest())
    dag = gb.build(["jolt"])
    task = dag.select(lambda graph, task: True)
    assert len(task) == 1, "Too many selfdeploy tasks found"
    task = task[0]
    if not task.is_available_remotely(cache=False):
        factory = LocalExecutorFactory()
        executor = LocalExecutor(factory, task, force_upload=True)
        executor.run(env)
    jolt_url = acache.location(task.artifacts[0])
    raise_error_if(not jolt_url, "Failed to deploy jolt to a remote cache")
    cacheUrl = config.get("http", "uri", config.get("cache", "uri", "") + "/files")
    substituteUrl = config.get("selfdeploy", "baseUri")
    if cacheUrl and substituteUrl:
        return task.identity, jolt_url.replace(cacheUrl, substituteUrl)
    return task.identity, jolt_url


def get_floating_version():
    identity, url = publish_artifact()
    return common_pb.Client(
        identity=identity,
        url=url,
        version=version.__version__,
    )


def get_pinned_version():
    return common_pb.Client(
        requirements=get_extra_dependencies(),
        version=version.__version__,
    )


def get_client():
    floating = config.getboolean("selfdeploy", "floating", False)
    if floating:
        return get_floating_version()
    return get_pinned_version()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: jolt/plugins/remote_execution/administration.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from jolt.plugins.remote_execution import scheduler_pb2 as jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2jolt/plugins/remote_execution/administration.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a-jolt/plugins/remote_execution/scheduler.proto\"\x13\n\x11ListBuildsRequest\"T\n\x12ListBuildsResponse\x12)\n\x06\x62uilds\x18\x01 \x03(\x0b\x32\x19.ListBuildsResponse.Build\x1a\x13\n\x05\x42uild\x12\n\n\x02id\x18\x01 \x01(\t\"\x14\n\x12ListWorkersRequest\"\x15\n\x13ListWorkersResponse2\xf9\x01\n\x0e\x41\x64ministration\x12\x38\n\x0b\x43\x61ncelBuild\x12\x13.CancelBuildRequest\x1a\x14.CancelBuildResponse\x12\x35\n\nListBuilds\x12\x12.ListBuildsRequest\x1a\x13.ListBuildsResponse\x12\x38\n\x0bListWorkers\x12\x13.ListWorkersRequest\x1a\x14.ListWorkersResponse\x12<\n\nReschedule\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.EmptyB\x0eZ\x0cpkg/protocolb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'jolt.plugins.remote_execution.administration_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\014pkg/protocol'
  _LISTBUILDSREQUEST._serialized_start=130
  _LISTBUILDSREQUEST._serialized_end=149
  _LISTBUILDSRESPONSE._serialized_start=151
  _LISTBUILDSRESPONSE._serialized_end=235
  _LISTBUILDSRESPONSE_BUILD._serialized_start=216
  _LISTBUILDSRESPONSE_BUILD._serialized_end=235
  _LISTWORKERSREQUEST._serialized_start=237
  _LISTWORKERSREQUEST._serialized_end=257
  _LISTWORKERSRESPONSE._serialized_start=259
  _LISTWORKERSRESPONSE._serialized_end=280
  _ADMINISTRATION._serialized_start=283
  _ADMINISTRATION._serialized_end=532
# @@protoc_insertion_point(module_scope)



from threading import Thread, Event


class Timer(Thread):
    def __init__(self, interval, function, args=None, kwargs=None):
        Thread.__init__(self)
        self.interval = interval
        self.function = function
        self.args = args if args is not None else []
        self.kwargs = kwargs if kwargs is not None else {}
        self.finished = Event()

    def cancel(self):
        self.finished.set()

    def run(self):
        while True:
            self.finished.wait(self.interval)
            if not self.finished.is_set():
                self.function(*self.args, **self.kwargs)
            else:
                break



from concurrent.futures import ThreadPoolExecutor, as_completed, Future
import copy
import os
import queue

from jolt import common_pb2 as common_pb
from jolt import config
from jolt import hooks
from jolt import log
from jolt import utils
from jolt import tools
from jolt.error import raise_error
from jolt.error import raise_task_error
from jolt.error import raise_task_error_if
from jolt.graph import PruneStrategy
from jolt.manifest import ManifestExtension
from jolt.manifest import ManifestExtensionRegistry
from jolt.options import JoltOptions
from jolt.timer import Timer


class JoltEnvironment(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class TaskQueue(object):
    def __init__(self, strategy, cache, session):
        self.futures = {}
        self.strategy = strategy
        self.cache = cache
        self.session = session
        self.duration_acc = utils.duration_diff(0)
        self._aborted = False
        self._timer = Timer(60, self._log_task_running_time)
        self._timer.start()

    def _log_task_running_time(self):
        for future in self.futures:
            self.futures[future].task.log_running_time()

    def submit(self, task):
        if self._aborted:
            return None

        env = JoltEnvironment(cache=self.cache)
        executor = self.strategy.create_executor(self.session, task)
        raise_task_error_if(
            not executor, task,
            "no executor can execute the task; "
            "requesting a distributed network build without proper configuration?")

        task.set_in_progress()
        future = executor.submit(env)
        self.futures[future] = executor
        return future

    def wait(self):
        for future in as_completed(self.futures):
            task = self.futures[future].task
            try:
                future.result()
            except Exception as error:
                return task, error
            finally:
                self.duration_acc += task.duration_running or 0
                del self.futures[future]
            return task, None
        return None, None

    def abort(self):
        self._aborted = True
        for future, executor in self.futures.items():
            executor.cancel()
            future.cancel()
        if len(self.futures):
            log.info("Waiting for tasks to finish, please be patient")
        self.strategy.executors.shutdown()
        self._timer.cancel()

    def shutdown(self):
        self._timer.cancel()

    def is_aborted(self):
        return self._aborted

    def in_progress(self, task):
        return task in self.futures.values()

    def empty(self):
        return len(self.futures) == 0


class Executor(object):
    def __init__(self, factory):
        self.factory = factory
        self._status = None

    def submit(self, env):
        return self.factory.submit(self, env)

    def cancel(self):
        pass

    def is_aborted(self):
        return self.factory.is_aborted()

    def run(self, env):
        pass


class LocalExecutor(Executor):
    def __init__(self, factory, task, force_upload=False, force_build=False):
        super().__init__(factory)
        self.task = task
        self.force_build = force_build
        self.force_upload = force_upload

    def _run(self, env, task):
        if self.is_aborted():
            return
        try:
            with hooks.task_run(task):
                self.task.run(
                    env.cache,
                    force_build=self.force_build,
                    force_upload=self.force_upload)

        except Exception as e:
            log.exception(e, error=False)
            if not task.is_unstable:
                self.task.raise_for_status(log_error=getattr(env, "worker", False))
                raise e

        return task

    def get_all_extensions(self, task):
        extensions = copy.copy(task.extensions)
        for ext in extensions:
            extensions.extend(self.get_all_extensions(ext))
        return extensions

    def run(self, env):
        tasks = [self.task] + self.get_all_extensions(self.task)
        for task in tasks:
            task.queued()

        self._run(env, self.task)


class NetworkExecutor(Executor):
    pass


class SkipTask(Executor):
    def __init__(self, factory, task, *args, **kwargs):
        super().__init__(factory, *args, **kwargs)
        self.task = task

    def run(self, env):
        self.task.skipped()
        for ext in self.task.extensions:
            ext.skipped()
        return self.task


class Downloader(Executor):
    def __init__(self, factory, task, *args, **kwargs):
        super().__init__(factory, *args, **kwargs)
        self.task = task

    def _download(self, env, task):
        if self.is_aborted():
            return
        if not task.is_downloadable():
            return
        try:
            task.started_download()
            raise_task_error_if(
                not task.download(persistent_only=True),
                task, "Failed to download task artifact")
        except Exception as e:
            with task.task.report() as report:
                report.add_exception(e)
            task.failed_download()
            raise e
        else:
            task.finished_download()

    def run(self, env):
        self._download(env, self.task)
        for ext in self.task.extensions:
            self._download(env, ext)
        return self.task


class Uploader(Executor):
    def __init__(self, factory, task, *args, **kwargs):
        super().__init__(factory, *args, **kwargs)
        self.task = task

    def _upload(self, env, task):
        if self.is_aborted():
            return
        try:
            task.started_upload()
            raise_task_error_if(
                not task.upload(persistent_only=True),
                task, "Failed to upload task artifact")
        except Exception as e:
            with task.task.report() as report:
                report.add_exception(e)
            task.failed_upload()
            raise e
        else:
            task.finished_upload()

    def run(self, env):
        self._upload(env, self.task)
        for ext in self.task.extensions:
            self._upload(env, ext)

        return self.task


@utils.Singleton
class ExecutorRegistry(object):
    executor_factories = []
    extension_factories = []

    def __init__(self, options=None):
        self._options = options or JoltOptions()
        self._factories = [factory(self._options) for factory in self.__class__.executor_factories]
        self._local_factory = LocalExecutorFactory(self._options)
        self._concurrent_factory = ConcurrentLocalExecutorFactory(self._options)
        self._extensions = [factory().create() for factory in self.__class__.extension_factories]

    def shutdown(self):
        for factory in self._factories:
            factory.shutdown()
        self._local_factory.shutdown()
        self._concurrent_factory.shutdown()

    def create_session(self, graph):
        return {factory: factory.create_session(graph) for factory in self._factories}

    def create_skipper(self, task):
        return SkipTask(self._concurrent_factory, task)

    def create_downloader(self, task):
        # TODO: Switch to concurrent factory once the progress bar can handle it
        return Downloader(self._concurrent_factory, task)

    def create_uploader(self, task):
        # TODO: Switch to concurrent factory once the progress bar can handle it
        return Uploader(self._concurrent_factory, task)

    def create_local(self, task, force=False):
        task.set_locally_executed()
        return self._local_factory.create(task, force=force)

    def create_network(self, session, task):
        for factory in self._factories:
            executor = factory.create(session[factory], task)
            if executor is not None:
                task.set_remotely_executed()
                return executor
        return self.create_local(task)

    def get_network_parameters(self, task):
        parameters = {}
        for extension in self._extensions:
            parameters.update(extension.get_parameters(task))
        return parameters


class NetworkExecutorExtensionFactory(object):
    @staticmethod
    def Register(cls):
        ExecutorRegistry.extension_factories.insert(0, cls)
        return cls

    def create(self):
        raise NotImplementedError()


class NetworkExecutorExtension(object):
    def get_parameters(self, task):
        return {}


class Job(object):
    def __init__(self, priority, future, executor, env):
        self.priority = priority
        self.future = future
        self.executor = executor
        self.env = env

    def __le__(self, o):
        return self.priority <= o.priority

    def __ge__(self, o):
        return self.priority >= o.priority

    def __lt__(self, o):
        return self.priority < o.priority

    def __gt__(self, o):
        return self.priority > o.priority

    def __eq__(self, o):
        return self.priority == o.priority


class ExecutorFactory(object):
    @staticmethod
    def Register(cls):
        ExecutorRegistry.executor_factories.insert(0, cls)
        return cls

    def __init__(self, options=None, max_workers=None):
        self.pool = ThreadPoolExecutor(max_workers=max_workers)
        self._aborted = False
        self._queue = queue.PriorityQueue()
        self._options = options or JoltOptions()

    def is_aborted(self):
        return self._aborted

    def is_keep_going(self):
        return self._options.keep_going

    def shutdown(self):
        self._aborted = True
        self.pool.shutdown()

    def create(self, task):
        raise NotImplementedError()

    def _run(self):
        job = self._queue.get(False)
        self._queue.task_done()
        try:
            if not self.is_aborted():
                job.executor.run(job.env)
        except KeyboardInterrupt as e:
            raise_error("Interrupted by user")
            self._aborted = True
            job.future.set_exception(e)
        except Exception as e:
            if not self.is_keep_going():
                self._aborted = True
            job.future.set_exception(e)
        else:
            job.future.set_result(job.executor)

    def submit(self, executor, env):
        future = Future()
        self._queue.put(Job(-executor.task.weight, future, executor, env))
        self.pool.submit(self._run)
        return future


class LocalExecutorFactory(ExecutorFactory):
    def __init__(self, options=None):
        max_workers = config.getint(
            "jolt", "parallel_tasks",
            os.getenv("JOLT_PARALLEL_TASKS", 1 if options is None else options.jobs))
        super().__init__(
            options=options,
            max_workers=max_workers)

    def create(self, task, force=False):
        return LocalExecutor(self, task, force_build=force)


class ConcurrentLocalExecutorFactory(ExecutorFactory):
    def __init__(self, options=None):
        max_workers = tools.Tools().thread_count()
        super().__init__(
            options=options,
            max_workers=max_workers)

    def create(self, task):
        raise NotImplementedError()


class NetworkExecutorFactory(ExecutorFactory):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def create(self, session, task):
        raise NotImplementedError()


class ExecutionStrategy(object):
    def create_executor(self, session, task):
        raise NotImplementedError()


class LocalStrategy(ExecutionStrategy, PruneStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, session, task):
        if task.is_alias():
            return self.executors.create_skipper(task)
        if not task.is_cacheable():
            return self.executors.create_local(task)
        if task.is_available_locally():
            return self.executors.create_skipper(task)
        if self.cache.download_enabled() and task.is_available_remotely():
            return self.executors.create_downloader(task)
        return self.executors.create_local(task)

    def should_prune_requirements(self, task):
        if task.is_alias() or not task.is_cacheable():
            return False
        if task.is_available_locally():
            return True
        if self.cache.download_enabled() and task.is_available_remotely():
            return True
        return False


class DownloadStrategy(ExecutionStrategy, PruneStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, session, task):
        if task.is_alias():
            return self.executors.create_skipper(task)
        if task.is_resource():
            return self.executors.create_local(task)
        if not task.is_cacheable():
            return self.executors.create_skipper(task)
        if task.is_available_locally():
            return self.executors.create_skipper(task)
        if self.cache.download_enabled() and task.is_available_remotely(cache=False):
            return self.executors.create_downloader(task)
        raise_task_error(task, "Task must be built first")

    def should_prune_requirements(self, task):
        return False


class DistributedStrategy(ExecutionStrategy, PruneStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, session, task):
        if task.is_alias():
            return self.executors.create_skipper(task)

        if task.is_resource():
            if task.deps_available_locally():
                return self.executors.create_local(task)
            else:
                return self.executors.create_skipper(task)

        if not task.is_cacheable():
            return self.executors.create_network(session, task)

        if not self.cache.upload_enabled():
            return self.executors.create_network(session, task)

        if not task.is_goal(with_extensions=False):
            task.disable_download()
        for extension in task.extensions:
            if not extension.is_goal(with_extensions=False):
                extension.disable_download()

        remote = task.is_available_remotely()
        if remote:
            if task.is_goal() and self.cache.download_enabled() and \
               not task.is_available_locally():
                return self.executors.create_downloader(task)
            return self.executors.create_skipper(task)
        else:
            if task.is_available_locally() and task.is_uploadable():
                return self.executors.create_uploader(task)
            if task.is_fast() and task.deps_available_locally():
                return self.executors.create_local(task, force=True)

        return self.executors.create_network(session, task)

    def should_prune_requirements(self, task):
        if task.is_alias() or not task.is_cacheable():
            return False
        if task.is_available_remotely():
            return True
        return False


class WorkerStrategy(ExecutionStrategy, PruneStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, session, task):
        if task.is_resource():
            return self.executors.create_local(task)

        if task.is_alias():
            return self.executors.create_skipper(task)

        raise_task_error_if(
            not self.cache.upload_enabled(), task,
            "Artifact upload must be enabled for workers, fix configuration")

        if not task.is_cacheable():
            return self.executors.create_local(task)

        if task.is_available_locally():
            if task.is_goal() and not task.is_available_remotely():
                # Unpacked artifacts may become unpacked before we manage to upload.
                # To keep the implementation simple we take the easy road and rebuild
                # all artifacts that have not been unpacked, even if they are uploadable.
                if task.is_unpacked() and task.is_uploadable():
                    return self.executors.create_uploader(task)
                else:
                    return self.executors.create_local(task, force=True)
            return self.executors.create_skipper(task)

        if not self.cache.download_enabled():
            return self.executors.create_local(task)

        if task.is_available_remotely():
            return self.executors.create_downloader(task)
        elif not task.is_goal():
            raise_task_error(task, "Task artifact removed from global cache, cannot continue")

        return self.executors.create_local(task)

    def should_prune_requirements(self, task):
        if task.is_alias() or not task.is_cacheable():
            return False
        if task.is_available_locally():
            # Unpacked artifacts may become unpacked before we manage to upload.
            # To keep the implementation simple we take the easy road and rebuild
            # all artifacts that have not been unpacked, even if they are uploadable.
            if task.is_unpacked() and task.is_uploadable():
                return True
        if not task.is_goal() and task.task.selfsustained:
            return True
        return False


def get_exported_task_set(task):
    children = [task] + task.descendants
    for ext in task.extensions:
        children.extend(get_exported_task_set(ext))
    return list(set(children))


class TaskIdentityExtension(ManifestExtension):
    def export_manifest(self, manifest, tasks):
        # Generate a list of all tasks that must be evaluated
        # for inclusion in the manifest
        all_tasks = []
        for task in tasks:
            all_tasks += get_exported_task_set(task)
        all_tasks = list(set(all_tasks))

        for child in all_tasks:
            manifest_task = manifest.find_task(child.qualified_name)
            if manifest_task is None:
                manifest_task = manifest.create_task()
                manifest_task.name = child.qualified_name
            manifest_task.identity = child.identity
            manifest_task.instance = child.instance


ManifestExtensionRegistry.add(TaskIdentityExtension())


class TaskExportExtension(ManifestExtension):
    def export_manifest(self, manifest, tasks):
        short_task_names = set()

        # Generate a list of all tasks that must be evaluated
        # for inclusion in the manifest
        all_tasks = []
        for task in tasks:
            all_tasks += get_exported_task_set(task)
        all_tasks = list(set(all_tasks))

        # Add all tasks with export attributes to the manifest
        for child in all_tasks:
            manifest_task = manifest.find_task(child.qualified_name)
            if manifest_task is None:
                manifest_task = manifest.create_task()
                manifest_task.name = child.qualified_name
            for key, export in child.task._get_export_objects().items():
                attrib = manifest_task.create_attribute()
                attrib.name = key
                attrib.value = export.export(child.task)
            short_task_names.add(child.name)

        # Figure out if any task with an overridden default parameter
        # value was included in the manifest. If so, add info about it.
        default_task_names = set()
        for task in all_tasks:
            for task in task.options.default:
                short_name, _ = utils.parse_task_name(task)
                if short_name in short_task_names:
                    default_task_names.add(task)
        if default_task_names:
            build = manifest.create_build()
            for task in default_task_names:
                default = build.create_default()
                default.name = task


ManifestExtensionRegistry.add(TaskExportExtension())


def export_tasks(tasks):
    pb_tasks = {}

    for task in tasks:
        properties = []
        for key, export in task.task._get_export_objects().items():
            value = export.export(task.task)
            if value is not None:
                pb_attrib = common_pb.Property(key=key, value=str(value))
                properties.append(pb_attrib)

        platform = common_pb.Platform(
            properties=[
                common_pb.Property(key=key, value=value)
                for key, value in task.task.platform.items()
            ]
        )

        args = dict(
            identity=task.identity,
            instance=task.instance,
            taint=str(task.task.taint),
            name=task.short_qualified_name,
            platform=platform,
            properties=properties,
        )

        pb_tasks[task.short_qualified_name] = common_pb.Task(**args)

    return pb_tasks


def export_task_default_params(tasks):
    default_task_names = {}

    for task in tasks:
        for task in task.options.default:
            short_name, params = utils.parse_task_name(task)
            if short_name in default_task_names:
                default_task_names[short_name].update(params)
            else:
                default_task_names[short_name] = params

    return [
        utils.format_task_name(name, params)
        for name, params in default_task_names.items()
    ]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: jolt/plugins/remote_execution/log.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from jolt import common_pb2 as jolt_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\'jolt/plugins/remote_execution/log.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x11jolt/common.proto\"s\n\x0eReadLogRequest\x12\n\n\x02id\x18\x01 \x01(\t\x12*\n\x06\x62\x65\x66ore\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12)\n\x05\x61\x66ter\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"9\n\x0fReadLogResponse\x12\n\n\x02id\x18\x01 \x01(\t\x12\x1a\n\x08loglines\x18\x02 \x03(\x0b\x32\x08.LogLine2:\n\x08LogStash\x12.\n\x07ReadLog\x12\x0f.ReadLogRequest\x1a\x10.ReadLogResponse0\x01\x42\x0eZ\x0cpkg/protocolb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'jolt.plugins.remote_execution.log_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\014pkg/protocol'
  _READLOGREQUEST._serialized_start=95
  _READLOGREQUEST._serialized_end=210
  _READLOGRESPONSE._serialized_start=212
  _READLOGRESPONSE._serialized_end=269
  _LOGSTASH._serialized_start=271
  _LOGSTASH._serialized_end=329
# @@protoc_insertion_point(module_scope)
